use std::collections::BTreeMap;

use nom::{
    branch::alt,
    bytes::complete::tag,
    character::complete::{digit1, newline, space0},
    combinator::{map, map_res, opt},
    multi::{many1, separated_list1},
    sequence::{pair, preceded, terminated},
    IResult,
};

pub mod inputs;
mod test;

type Int = u32;

#[derive(Debug, PartialOrd, Ord, PartialEq, Eq, Clone)]
enum Colour {
    Blue,
    Red,
    Green,
}

impl Colour {
    fn limit(&self) -> Int {
        match self {
            Colour::Blue => 14,
            Colour::Red => 12,
            Colour::Green => 13,
        }
    }
}

#[derive(Debug)]
struct Pick {
    colour: Colour,
    quantity: Int,
}

type Hand = Vec<Pick>;

#[derive(Debug)]
struct Game {
    id: Int,
    hands: Vec<Hand>,
}

fn parse_colour(input: &str) -> IResult<&str, Colour> {
    alt((
        map(tag("blue"), |_| Colour::Blue),
        map(tag("red"), |_| Colour::Red),
        map(tag("green"), |_| Colour::Green),
    ))(input)
}

fn parse_number(input: &str) -> IResult<&str, Int> {
    map_res(digit1, |d: &str| d.parse::<Int>())(input)
}

fn parse_pick(input: &str) -> IResult<&str, Pick> {
    let (input, quantity) = parse_number(input)?;
    let (input, _) = space0(input)?;
    let (input, colour) = parse_colour(input)?;
    let pick = Pick { colour, quantity };
    Ok((input, pick))
}

fn parse_hand(input: &str) -> IResult<&str, Hand> {
    separated_list1(pair(tag(","), space0), parse_pick)(input)
}

fn parse_game(input: &str) -> IResult<&str, Game> {
    let (input, id) = preceded(pair(tag("Game"), space0), parse_number)(input)?;
    let (input, _) = pair(tag(":"), space0)(input)?;
    let (input, hands) = separated_list1(pair(tag(";"), space0), parse_hand)(input)?;
    let game = Game { id, hands };
    Ok((input, game))
}

fn parse_input(input: &str) -> IResult<&str, Vec<Game>> {
    terminated(many1(terminated(parse_game, newline)), opt(newline))(input)
}

pub fn solution(input: &str) -> Result<(Int, Int), String> {
    let (_, output) = parse_input(input).map_err(|e| e.to_string())?;
    let (a, b) = output
        .into_iter()
        .map(|g| {
            let mut limits: BTreeMap<Colour, Int> =
                vec![(Colour::Red, 0), (Colour::Green, 0), (Colour::Blue, 0)]
                    .into_iter()
                    .collect();
            for h in &g.hands {
                for p in h {
                    let xp = limits.entry(p.colour.clone()).or_insert(0);
                    *xp = p.quantity.max(*xp);
                }
            }
            let valid = g
                .hands
                .into_iter()
                .all(|h| h.into_iter().all(|p| p.quantity <= p.colour.limit()));
            let power: Int = limits.into_values().product();
            if valid {
                (g.id, power)
            } else {
                (0, power)
            }
        })
        .fold((0, 0), |(a, b), (c, d)| (a + c, b + d));
    Ok((a, b))
}
